#include <kernel.h>
#include <printf.h>
#include <kdata.h>

/* Buffer pool management */
/*********************************************************************
The high-level interface is through bread() and bfree().
Bread() is given a device and block number, and a rewrite flag.  If
rewrite is 0, the block is actually read if it is not already in the
buffer pool. If rewrite is set, it is assumed that the caller plans to
rewrite the entire contents of the block, and it will not be read in,
but only have a buffer named after it.

Bfree() is given a buffer pointer and a dirty flag.  If the dirty flag
is 0, the buffer is made available for further use.  If the flag is 1,
the buffer is marked "dirty", and it will eventually be written out to
disk.  If the flag is 2, it will be immediately written out.

Zerobuf() returns a buffer of zeroes not belonging to any device.  It
must be bfree'd after use, and must not be dirty. It is used when a
read() wants to read an unallocated block of a file.

Bufsync() write outs all dirty blocks.

Note that a pointer to a buffer structure is the same as a pointer to
the data.  This is very important.
**********************************************************************/

uint16_t bufclock;         /* Time-stamp counter for LRU */

char *bread (uint8_t dev, blkno_t blk, bool rewrite)
{
    register bufptr bp;

    if (bp = bfind (dev, blk))
    {
        if (bp->bf_busy)
            panic ("want busy block");
        goto done;
    }
    bp = freebuf();
    bp->bf_dev = dev;
    bp->bf_blk = blk;

    /* If rewrite is set, we are about to write over the entire block,
       so we don't need the previous contents */
    if(!rewrite){
        if(bdread(bp) == -1){
            udata.u_error = EIO;
            return (NULL);
        }
    }

done:
    bp->bf_busy = 1;
    bp->bf_time = ++bufclock;  /* Time stamp it */
    return (bp->bf_data);
}


void brelse(void *bp)
{
    bfree((bufptr)bp, 0);
}


void bawrite(void * bp)
{
    bfree((bufptr)bp, 1);
}


int bfree(bufptr bp, uint8_t dirty) /* dirty: 0=clean, 1=dirty (write back), 2=dirty+immediate write */
{
    if(dirty)
        bp->bf_dirty = true;
    bp->bf_busy = false;

    if(dirty > 1){ // immediate writeback
        if (bdwrite(bp) == -1)
            udata.u_error = EIO;
        bp->bf_dirty = false;
        return -1;
    }
    return 0;
}


/* This returns a busy block not belonging to any device, with
 * garbage contents.  It is essentially a malloc for the kernel.
 * Free it with brelse()!
 */
void *tmpbuf(void)
{
    bufptr bp;

    bp = freebuf();
    bp->bf_dev = NO_DEVICE;
    bp->bf_busy = true;
    bp->bf_time = ++bufclock;   /* Time stamp it */
    return bp->bf_data;
}


void *zerobuf(void)
{
    void *b;

    b = tmpbuf();
    memset(b, 0, 512);

    return b;
}


void bufsync(void)
{
    bufptr bp;

    for (bp=bufpool; bp < bufpool+NBUFS; ++bp){
        if ((bp->bf_dev != NO_DEVICE) && bp->bf_dirty){
            bdwrite(bp);
            if (!bp->bf_busy)
                bp->bf_dirty = false;
        }
    }
}

bufptr bfind(uint8_t dev, blkno_t blk)
{
    bufptr bp;

    for (bp=bufpool; bp < bufpool+NBUFS; ++bp){
        if (bp->bf_dev == dev && bp->bf_blk == blk)
            return bp;
    }
    return NULL;
}


bufptr freebuf(void)
{
    bufptr bp;
    bufptr oldest;
    int16_t oldtime;

    /* Try to find a non-busy buffer and write out the data if it is dirty */
    oldest = NULL;
    oldtime = 0;
    for (bp=bufpool; bp < bufpool+NBUFS; ++bp) {
        if (bufclock - bp->bf_time >= oldtime && !bp->bf_busy) {
            oldest = bp;
            oldtime = bufclock - bp->bf_time;
        }
    }
    if(!oldest)
        panic("no free buffers");

    if (oldest->bf_dirty) {
        if(bdwrite(oldest) == -1)
            udata.u_error = EIO;
        oldest->bf_dirty = false;
    }
    return oldest;
}
        

void bufinit(void)
{
    bufptr bp;

    for (bp=bufpool; bp < bufpool+NBUFS; ++bp){
        bp->bf_dev = NO_DEVICE;
        bp->bf_busy = false;
    }
}


void bufdump (void)
{
    bufptr j;

    kprintf("\ndev\tblock\tdirty\tbusy\ttime clock %d\n", bufclock);
    for(j=bufpool; j < bufpool+NBUFS; ++j)
        kprintf ("%d\t%u\t%d\t%d\t%u\n", j->bf_dev, j->bf_blk, j->bf_dirty, j->bf_busy, j->bf_time);
}


/*********************************************************************
Bdread() and bdwrite() are the block device interface routines.  They
are given a buffer pointer, which contains the device, block number,
and data location.  They basically validate the device and vector the
call.

Cdread() and cdwrite are the same for character (or "raw") devices,
and are handed a device number.  Udata.u_base, count, and offset have
the rest of the data.
**********************************************************************/

/*********************************************************************
The device driver read and write routines now have only two arguments,
minor and rawflag.  If rawflag is zero, a single block is desired, and
the necessary data can be found in udata.u_buf.  Otherwise, a "raw" or
character read is desired, and udata.u_offset, udata.u_count, and
udata.u_base should be consulted instead.
Any device other than a disk will have only raw access.
**********************************************************************/

int bdread(bufptr bp)
{
    if(!validdev(bp->bf_dev))
        panic("bdread: invalid dev");

    udata.u_buf = bp;
    return ((*dev_tab[bp->bf_dev].dev_read)(dev_tab[bp->bf_dev].minor, 0));
}


int bdwrite(bufptr bp)
{
   if(!validdev(bp->bf_dev))
       panic("bdwrite: invalid dev");

    udata.u_buf = bp;
    return ((*dev_tab[bp->bf_dev].dev_write)(dev_tab[bp->bf_dev].minor, 0));
}

int cdread(uint8_t dev)
{
    if(!validdev(dev))
        panic("cdread: invalid dev");
    return ((*dev_tab[dev].dev_read)(dev_tab[dev].minor, 1));
}

int cdwrite(uint8_t dev)
{
    if(!validdev(dev))
        panic("cdwrite: invalid dev");
    return ((*dev_tab[dev].dev_write)(dev_tab[dev].minor, 1));
}

// WRS: swapread(), swapwrite() removed.

int d_open(uint8_t dev)
{
    if(!validdev(dev))
        return -1;
    return ((*dev_tab[dev].dev_open)(dev_tab[dev].minor));
}


int d_close(uint8_t dev)
{
    if(!validdev(dev))
        panic("d_close: bad device");
    return (*dev_tab[dev].dev_close)(dev_tab[dev].minor);
}

int d_ioctl(uint8_t dev, uint16_t request, char *data)
{
    if(!validdev(dev)){
        udata.u_error = ENXIO;
        return -1;
    }

    if ((*dev_tab[dev].dev_ioctl)(dev_tab[dev].minor, request, data)){
        if(!udata.u_error) // maybe the ioctl routine might set this?
            udata.u_error = EINVAL;
        return -1;
    }

    return 0;
}

/*********************************************************************
             Character queue management routines
**********************************************************************/

/* add something to the tail of the queue. */
bool insq(struct s_queue *q, char c)
{
    bool r;
    __critical {
        if (q->q_count == q->q_size)
            r = false; // no space left :(
        else{
            *(q->q_tail) = c;
            ++q->q_count;
            if (++q->q_tail >= q->q_base + q->q_size)
                q->q_tail = q->q_base;
            r = true;
        }
    }
    return r;
}


/* Remove something from the head of the queue. */
bool remq(struct s_queue *q, char *cp)
{
    bool r;
    __critical {
        if(!q->q_count)
            r = false;
        else{
            *cp = *(q->q_head);
            --q->q_count;
            if (++q->q_head >= q->q_base + q->q_size)
                q->q_head = q->q_base;
            r = true;
        }
    }
    return r;
}


/* Clear the queue to empty conditions.  (UZI280 addition) */
void clrq(struct s_queue *q)
{
    __critical {
        q->q_head = q->q_tail = q->q_base;
        q->q_count = 0;
    }
}


/* Remove something from the tail; the most recently added char. */
bool uninsq(struct s_queue *q, char *cp)
{
    bool r;
    __critical {
        if(!q->q_count)
            r = false;
        else {
            --q->q_count;
            if (--q->q_tail < q->q_base)
                q->q_tail = q->q_base + q->q_size - 1;
            *cp = *(q->q_tail);
            r = true;
        }
    }
    return r;
}

// WRS: this isn't used
// /* Returns true if the queue has more characters than its wakeup number
//  */
// bool fullq(struct s_queue *q)
// {
//     if (q->q_count > q->q_wakeup) // WRS: shouldn't this be >= ?
//         return true;
//     else
//         return false;
// }
