#include <kernel.h>
#include <kdata.h>
#include <printf.h>
#include <stdbool.h>
#include <devtty.h>

#undef  DEBUG            /* UNdefine to delete debug code sequences */

struct tty_data ttydata[NUM_DEV_TTY+1];        /* ttydata[0] is not used */

struct tty_data ttydflt = {
    0,         0,         '\b',       CTRL('x'), DFLT_MODE,
    CTRL('c'), CTRL('\\'), CTRL('q'), CTRL('s'), CTRL('d'), 0
};

char tbuf1[TTYSIZ];
char tbuf2[TTYSIZ];

struct  s_queue  ttyinq[NUM_DEV_TTY+1] = {       /* ttyinq[0] is never used */
    {   NULL,    NULL,    NULL,    0,        0,       0    },
    {   tbuf1,   tbuf1,   tbuf1,   TTYSIZ,   0,   TTYSIZ/2 },
    {   tbuf2,   tbuf2,   tbuf2,   TTYSIZ,   0,   TTYSIZ/2 }
};

bool stopflag[NUM_DEV_TTY+1];   // Flag for ^S/^Q
bool flshflag[NUM_DEV_TTY+1];   // Flag for ^O

int tty_read(uint8_t minor, uint8_t rawflag)
{
    uint16_t nread;
    char c;
    
    rawflag; // shut up compiler

    /* Minor == 0 means that it is the controlling tty of the process */
    if(!minor)
        minor = udata.u_ptab->p_tty;
    if(!udata.u_ptab->p_tty)
        udata.u_ptab->p_tty = minor;

    if ((minor < 1) || (minor > NUM_DEV_TTY+1)) {
        udata.u_error = ENODEV;
        return (-1);
    }

    nread = 0;
    while (nread < udata.u_count) {
        for (;;) {
            if (remq (&ttyinq[minor], &c)) {
                if (udata.u_sysio)
                    *udata.u_base = c;
                else{
                    // kprintf("uputc('%c', 0x%04x)", c, udata.u_base);
                    uputc(c, udata.u_base);
                }
                break;
            }
            psleep(&ttyinq[minor]);
            if (udata.u_cursig || udata.u_ptab->p_pending) {     /* messy */
                udata.u_error = EINTR;
                return (-1);
            }
        }

        ++nread;

        /* In raw or cbreak mode, return after one char */
        if (ttydata[minor].t_flags & (RAW|CBREAK))
            break;

        if ((nread == 1) && (c == ttydata[minor].t_eof))   /* ^D */
            return(0);

        if (c == '\n')
            break;

        ++udata.u_base;
    } 

    return nread;
}


int tty_write(uint8_t minor, uint8_t rawflag)
{
    int towrite;
    uint8_t c;

    rawflag; // shut up compiler

    /* Minor == 0 means that it is the controlling tty of the process */
    if(!minor)
        minor = udata.u_ptab->p_tty;
    if(!udata.u_ptab->p_tty)
        udata.u_ptab->p_tty = minor;

    if ((minor < 1) || (minor > NUM_DEV_TTY+1)) {
        udata.u_error = ENODEV;
        return (-1);
    }

    towrite = udata.u_count;

    while (udata.u_count-- != 0){
        for (;;){       /* Wait on the ^S/^Q flag */
            if(!stopflag[minor])    
                break;
            psleep(&stopflag[minor]);
            if(udata.u_cursig || udata.u_ptab->p_pending){  /* messy */
                udata.u_error = EINTR;
                return (-1);
            }
        }

        if(!flshflag[minor]){
            if (udata.u_sysio)
                c = *udata.u_base;
            else
                c = ugetc(udata.u_base);

            if (c == '\n' && (ttydata[minor].t_flags & CRMOD)){
                tty_putc_wait(minor, '\r');
            }
            tty_putc_wait(minor, c);
        }
        ++udata.u_base;
    }

    return towrite;
}

int tty_open(uint8_t minor)
{
    /* Minor == 0 means that it is the controlling tty of the process */
    if(!minor)
        minor = udata.u_ptab->p_tty;

    /* If there is no controlling tty for the process, establish it */
    if(!udata.u_ptab->p_tty)
        udata.u_ptab->p_tty = minor;

    /* Initialize the ttydata */
    memcpy(&ttydata[minor], &ttydflt, sizeof (struct tty_data));
    return (0);
}


int tty_close(uint8_t minor)
{
    /* If we are closing the controlling tty, make note */
    if (minor == udata.u_ptab->p_tty)
        udata.u_ptab->p_tty = 0;
    return (0);
}

int tty_ioctl (uint8_t minor, uint16_t request, char *data)           /* Data in User Space */
{
    if(!minor)
        minor = udata.u_ptab->p_tty;
    if((minor < 1) || (minor > NUM_DEV_TTY+1)) {
        udata.u_error = ENODEV;
        return (-1);
    }
    switch (request)
    {
        case TIOCGETP:
            uput(&ttydata[minor], data, 6);
            break;
        case TIOCSETP:
            uget(data, &ttydata[minor], 6);
            break;
        case TIOCGETC:
            uput(&ttydata[minor].t_intr, data, 5);
            break;
        case TIOCSETC:
            uget(data, &ttydata[minor].t_intr, 5);
            break;
        case TIOCSETN:
            uput(&ttyinq[minor].q_count, data, 2);
            break;
        case TIOCFLUSH:
            clrq(&ttyinq[minor]);
            break;
        case TIOCTLSET:
            ttydata[minor].ctl_char = 1;
            break;
        case TIOCTLRES:
            ttydata[minor].ctl_char = 0;
            break;
        default:
            udata.u_error = EINVAL;
            return (-1);
    }
    return (0);
}


/* This routine processes a character in response to an interrupt.  It
 * adds the character to the tty input queue, echoing and processing
 * backspace and carriage return.  If the queue contains a full line,
 * it wakes up anything waiting on it.  If it is totally full, it beeps
 * at the user.
 * UZI180 - This routine is called from the raw Hardware read routine,
 * either interrupt or polled, to process the input character.  HFB
 */

void tty_inproc(uint8_t minor, char c)
{
    char  oc;
    struct tty_data *td;
    int   mode;

    td = &ttydata[minor];
    mode = td->t_flags & (RAW|CBREAK|COOKED);

    if (mode != RAW)
        c &= 0x7f;                /* Strip off parity */
    if (!c)
        return;                   /* Simply quit if Null character */

#ifdef DEBUG
    if (c == 0x1a)                /* ^Z */
        idump();                  /*   (For debugging) */
    if (c == 0x01) /* ^A */
        trap_monitor();
#endif

    if (td->ctl_char == 0)         /* Don't parse ctl chars if Non-0 */
    {
        if ((mode & RAW) == 0)       /* if mode == COOKED or CBREAK */
        {
            if (c == '\r' && (td->t_flags & CRMOD))
                c = '\n';

            if (c == td->t_intr) {                   /* ^C */
                sgrpsig(minor, SIGINT);
                clrq(&ttyinq[minor]);
                stopflag[minor] = flshflag[minor] = false;
                return;
            }
            else if (c == td->t_quit) {              /* ^\ */
                sgrpsig(minor, SIGQUIT);
                clrq(&ttyinq[minor]);
                stopflag[minor] = flshflag[minor] = false;
                return;
            }
            else if (c == '\017') {                  /* ^O */
                flshflag[minor] = !flshflag[minor];
                return;
            }
            else if (c == td->t_stop) {              /* ^S */
                stopflag[minor] = true;
                return;
            }
            else if (c == td->t_start) {             /* ^Q */
                stopflag[minor] = false;
                wakeup (&stopflag[minor]);
                return;
            }
        }

        if(mode == COOKED)
        {
            if (c == 0x7f) /* handle DELETE as backspace */
                c = td->t_erase;

            if (c == td->t_erase)
            {
                if (uninsq(&ttyinq[minor], &oc))
                {
                    if (oc == '\n')
                        insq(&ttyinq[minor], oc);  /* Don't erase past nl */
                    else
                    {
                        tty_echo(minor, '\b');
                        tty_echo(minor, ' ');
                        tty_echo(minor, '\b');
                    }
                }
                return;
            }
            else if (c == td->t_kill)
            {
                while (uninsq(&ttyinq[minor], &oc))
                {
                    if (oc == '\n')
                    {
                        insq(&ttyinq[minor], oc);  /* Don't erase past nl */
                        break;
                    }
                    tty_echo(minor, '\b');
                    tty_echo(minor, ' ');
                    tty_echo(minor, '\b');
                }
                return;
            }
        }
    }

    /* All modes come here */
    if (c == '\n' && (td->t_flags & CRMOD))
        tty_echo(minor, '\r');

    if (insq (&ttyinq[minor], c))
        tty_echo(minor, c);
    else
        tty_putc_wait(minor, '\007');        /* Beep if no more room */

    if ((mode != COOKED) || ((c == '\n' || c == td->t_eof)
                && (td->ctl_char == 0)))                 /* ^D */
        wakeup(&ttyinq[minor]);
}

// called when a UART transmitter is ready for the next character
void tty_outproc(uint8_t minor)
{
    wakeup(&ttydata[minor]);
}

void tty_echo(uint8_t minor, char c)
{
    if (ttydata[minor].t_flags & ECHO)
        tty_putc_wait(minor, c);
}

// this is used by sdcc stdlib, assumes system console is on tty1
void putchar(char c)
{
    tty_putc(1, c);
    // handle CRLF
    if(c=='\n')
        tty_putc(1, '\r');
}

void tty_putc_wait(uint8_t minor, char c)
{
    if(!udata.u_ininterrupt){
        if(!tty_writeready(minor)){
            psleep(&ttydata[minor]);
        }
    }
    tty_putc(minor, c);
}
