#include <kernel.h>
#include <kdata.h>
#include <printf.h>
#include <timer.h>

uint16_t psize(inoptr ino)
{
    return (512*ino->c_node.i_size.o_blkno+ino->c_node.i_size.o_offset);
}

void addoff(off_t *ofptr, int16_t amount)
{
    int16_t temp;

    if (amount >= 0) {
        // positive
        temp = amount % 512;
        ofptr->o_offset += temp;
        if (ofptr->o_offset >= 512)
        {
            ofptr->o_offset -= 512;
            ++ofptr->o_blkno;
        }
        temp = amount / 512;
        ofptr->o_blkno += temp;
    } else {
        // negative
        temp = (-amount) % 512;
        ofptr->o_offset -= temp;
        if (ofptr->o_offset < 0)
        {
            ofptr->o_offset += 512;
            --ofptr->o_blkno;
        }
        temp = (-amount) / 512;
        ofptr->o_blkno -= temp;
    }
}

/* Writei (and readi) need more i/o error handling */
void readi(inoptr ino)
{
    uint16_t amount;
    uint16_t toread;
    blkno_t pblk;
    char *bp;
    uint8_t dev;
    bool ispipe;

    dev = ino->c_dev;
    ispipe = false;
    switch (getmode (ino))
    {
        case F_DIR:
        case F_REG:

            /* See if end of file will limit read */

            if (ino->c_node.i_size.o_blkno < udata.u_offset.o_blkno) {
                udata.u_count = 0;
            } else if ((ino->c_node.i_size.o_blkno == udata.u_offset.o_blkno) &&
                    (ino->c_node.i_size.o_offset < udata.u_offset.o_offset)) {
                udata.u_count = 0;
            } else {
                udata.u_count = ino->c_node.i_size.o_blkno - udata.u_offset.o_blkno >= 64 ?
                    udata.u_count : min(udata.u_count,
                            512 * (ino->c_node.i_size.o_blkno - udata.u_offset.o_blkno) +
                            (ino->c_node.i_size.o_offset - udata.u_offset.o_offset));
            }
            toread = udata.u_count;
            goto loop;

        case F_PIPE:
            ispipe = true;
            while (psize(ino) == 0)
            {
                if (ino->c_refs == 1)      /* No writers */
                    break;
                /* Sleep if empty pipe */
                psleep (ino);
            }
            toread = udata.u_count = min (udata.u_count, psize(ino));
            goto loop;

        case F_BDEV:
            toread = udata.u_count;
            dev = *(ino->c_node.i_addr);

loop:
            while (toread)
            {
                if ((pblk = bmap (ino, udata.u_offset.o_blkno, 1)) != NULLBLK)
                    bp = bread (dev, pblk, 0);
                else
                    bp = zerobuf();

                amount = min (toread, 512 - udata.u_offset.o_offset);

                /* if System is in context or destination is in Common memory
                 * (F000-FFFF is always in context), simply copy the data.
                 * Otherwise perform Interbank move (uput) to User bank.
                 */
                if (udata.u_sysio || (udata.u_base >= (char *)PROGTOP)){
                    memcpy(udata.u_base, bp+udata.u_offset.o_offset, amount);
                }else{
                    uput(bp+udata.u_offset.o_offset, udata.u_base, amount);
                }
                brelse (bp);

                udata.u_base += amount;
                addoff(&udata.u_offset, amount);
                if (ispipe && udata.u_offset.o_blkno >= 18)
                    udata.u_offset.o_blkno = 0;
                toread -= amount;
                if (ispipe)
                {
                    addoff (&(ino->c_node.i_size), -amount);
                    wakeup (ino);
                }
            }
            break;

        case F_CDEV:
            udata.u_count = cdread (ino->c_node.i_addr[0]);

            if (udata.u_count != -1)
                addoff (&udata.u_offset, udata.u_count);
            break;

        default:
            udata.u_error = ENODEV;
    }
}



void writei(inoptr ino)
{
    uint16_t amount;
    uint16_t towrite;
    char *bp;
    bool ispipe;
    blkno_t pblk;
    uint8_t dev;

    dev = ino->c_dev;

    switch (getmode (ino))
    {

        case F_BDEV:
            dev = *(ino->c_node.i_addr);
        case F_DIR:
        case F_REG:
            ispipe = false;
            towrite = udata.u_count;
            goto loop;

        case F_PIPE:
            ispipe = true;
            while ((towrite = udata.u_count) > (16*512) - psize(ino))
            {
                if (ino->c_refs == 1)        /* No readers */
                {
                    udata.u_count = -1;
                    udata.u_error = EPIPE;
                    ssig (udata.u_ptab, SIGPIPE);
                    return;
                }
                /* Sleep if empty pipe */
                psleep (ino);
            }
            /* Sleep if empty pipe */
            goto loop;

loop:

            while (towrite) {
                amount = min (towrite, 512 - udata.u_offset.o_offset);

                if ((pblk = bmap (ino, udata.u_offset.o_blkno, 0)) == NULLBLK)
                    break;                     /* No space to make more blocks */

                /* If we are writing an entire block, we don't care
                 * about its previous contents
                 */
                bp = bread (dev, pblk, (amount == 512));

                /* if System is in context or destination is in Common memory
                 * (F000-FFFF is always in context), simply copy the data.
                 * Otherwise perform Interbank move (uput) to User bank.
                 */
                if (udata.u_sysio || (udata.u_base >= (char *)PROGTOP)){
                    memcpy(bp+udata.u_offset.o_offset, udata.u_base, amount);
                }else{
                    uget(udata.u_base, bp+udata.u_offset.o_offset, amount);
                }

                bawrite (bp);

                udata.u_base += amount;
                addoff (&udata.u_offset, amount);
                if (ispipe)
                {
                    if (udata.u_offset.o_blkno >= 18)
                        udata.u_offset.o_blkno = 0;
                    addoff (&(ino->c_node.i_size), amount);
                    /* Wake up any readers */
                    wakeup (ino);
                }
                towrite -= amount;
            }

            /* Update size if file grew */
            if(!ispipe)
            {
                if (udata.u_offset.o_blkno > ino->c_node.i_size.o_blkno ||
                        (udata.u_offset.o_blkno == ino->c_node.i_size.o_blkno &&
                         udata.u_offset.o_offset > ino->c_node.i_size.o_offset))
                {
                    ino->c_node.i_size.o_blkno = udata.u_offset.o_blkno;
                    ino->c_node.i_size.o_offset = udata.u_offset.o_offset;
                    ino->c_dirty = 1;
                }
            }
            break;

        case F_CDEV:
            udata.u_count = cdwrite (ino->c_node.i_addr[0]);

            if (udata.u_count != -1)
                addoff (&udata.u_offset, udata.u_count);
            break;

        default:
            udata.u_error = ENODEV;
    }
}

int16_t doclose (uint8_t uindex)
{
    int8_t oftindex;
    inoptr ino;

    if(!(ino = getinode (uindex)))
        return (-1);

    oftindex = udata.u_files[uindex];

    if (isdevice (ino))
/*        && ino->c_refs == 1 && of_tab[oftindex].o_refs == 1) */
        d_close ((int)(ino->c_node.i_addr[0]));

    udata.u_files[uindex] = NO_FILE;
    oft_deref (oftindex);

    return (0);
}

inoptr rwsetup (bool is_read)
{
    inoptr ino;
    struct oft *oftp;

    udata.u_sysio = false;                     /* I/O to user data space */
    udata.u_base  = (char *)udata.u_argn1;     /* buf */
    udata.u_count = (uint16_t)udata.u_argn2;   /* nbytes */

    if((ino = getinode (udata.u_argn)) == NULLINODE){
        /* kprintf("[WRS: rwsetup(): getinode(%x) fails]", udata.u_argn); */
        return (NULLINODE);
    }

    oftp = of_tab + udata.u_files[udata.u_argn];
    if(oftp->o_access == (is_read ? O_WRONLY : O_RDONLY))
    {
        udata.u_error = EBADF;
        return (NULLINODE);
    }
    setftime (ino, is_read ? A_TIME : (A_TIME | M_TIME | C_TIME));

    /* Initialize u_offset from file pointer */
    udata.u_offset.o_blkno = oftp->o_ptr.o_blkno;
    udata.u_offset.o_offset = oftp->o_ptr.o_offset;

    return (ino);
}
