#undef  DEBUG         // turn this on to enable syscall tracing
#undef  DEBUGHARDER   // report calls to wakeup() that lead nowhere

#include <kernel.h>
#include <timer.h>
#include <kdata.h>
#include <printf.h>

/* psleep() puts a process to sleep on the given event.  If another
 * process is runnable, it swaps out the current one and starts the
 * new one.  Normally when psleep is called, the interrupts have
 * already been disabled.   An event of 0 means a pause(), while an
 * event equal to the process's own ptab address is a wait().
 */

void psleep(void *event)
{
#ifdef DEBUG
    kprintf("psleep(0x%x)", event); /* WRS */
#endif

    __critical {
        switch(udata.u_ptab->p_status){
            case P_RUNNING: // normal process
            case P_SLEEP:   // echo output from devtty happens while processes are still sleeping but in-context
                break;
            default:
                panic ("psleep: voodoo");
        }

        if(!event)
            udata.u_ptab->p_status = P_PAUSE;
        else if(event == (char *)udata.u_ptab)
            udata.u_ptab->p_status = P_WAIT;
        else
            udata.u_ptab->p_status = P_SLEEP;

        udata.u_ptab->p_wait = event;
        udata.u_ptab->p_waitno = ++waitno;
    }

    swapout();  /* Swap us out, and start another process */
    /* Swapout doesn't return in this context until we have been swapped back in, of course. */
}



/* wakeup() looks for any process waiting on the event,
 * and makes it runnable
 */

void wakeup(void *event)
{
    ptptr p;

#ifdef DEBUGHARDER
    kprintf("wakeup(0x%x)\n", event);
#endif

    __critical {
        for(p=ptab; p < ptab+maxproc; ++p)
        {
            if(p->p_status > P_RUNNING && p->p_wait == event) {
#ifdef DEBUG
                kprintf("wakeup: found proc 0x%x pid %d\n", p, p->p_pid);
#endif
                p->p_status = P_READY;
                p->p_wait = NULL;
            }
        }
    }
}



/* Getproc returns the process table pointer of a runnable process.
 * It is actually the scheduler.  If there are none, it loops.
 * This is the only time-wasting loop in the system.
 */

ptptr getproc_nextp = ptab;

ptptr getproc(void)
{
    ptptr haltafter;
#ifdef DEBUG
    kprintf("getproc start ... ");
#endif

    //for(pp=ptab; pp < ptab+maxproc; pp++)
    //    kprintf("ptab[0x%x]: pid=%d uid=%d status=%d, page=0x%x\n", pp, pp->p_pid, pp->p_uid, pp->p_status, pp->p_page);

    // yes please, interrupts on.
    // WRS: we could halt in here, after a full pass around?
    __asm
        ei
    __endasm;

    haltafter = getproc_nextp;

    for (;;) {
        getproc_nextp++;
        if (getproc_nextp >= ptab + maxproc) {
            getproc_nextp = ptab;
        }

        switch(getproc_nextp->p_status){
            case P_RUNNING:
                panic("getproc: extra running");
            case P_READY:
#ifdef DEBUG
                kprintf("[getproc returning %x pid=%d]\n", getproc_nextp, getproc_nextp->p_pid);
#endif
                return getproc_nextp;
        }
        if(getproc_nextp == haltafter){
            __asm
                halt
            __endasm;
        }
    }
}


/* Newproc fixes up the tables for the child of a fork but also for init
 * Call in the processes context!
 * This process MUST be run immediately (since it sets status P_RUNNING)
 */
void newproc(ptptr p) /* Passed New process table entry */
{
    uint8_t *j;

    __critical {
        /* Note that ptab_alloc clears most of the entry */
        /* calculate base page of process based on ptab table offset */
        udata.u_page = p->p_page;

        program_vectors(p->p_page); // set up vectors in new process and copy kernel code in common memory.

        p->p_status = P_RUNNING;

        p->p_pptr = udata.u_ptab;
        p->p_ignored = udata.u_ptab->p_ignored;
        p->p_tty = udata.u_ptab->p_tty;
        if(!p->p_tty)                /* If no tty, try tty of parent's parent */
            p->p_tty = udata.u_ptab->p_pptr->p_tty;
        p->p_uid = udata.u_ptab->p_uid;
        udata.u_ptab = p;

        memset(&udata.u_utime, 0, 4 * sizeof(uzitime_t)); /* Clear tick counters */

        rdtime(&udata.u_time);
        if(udata.u_cwd)
            i_ref(udata.u_cwd);
        udata.u_cursig = 0;
        udata.u_error = 0;

        /* Set default priority */
        p->p_priority = MAXTICKS;

        for(j=udata.u_files; j < (udata.u_files+UFTSIZE); ++j){
            if(*j != NO_FILE)
                ++of_tab[*j].o_refs;
        }
    }
}

/* This allocates a new process table slot, and fills in its
 * p_pid field with a unique number.
 */
uint16_t nextpid = 0;
ptptr ptab_alloc(void)
{
    ptptr p, newp;

    newp = NULL;

    __critical {
        for(p=ptab; p<ptab+maxproc; p++)
            if(p->p_status == P_EMPTY){
                newp = p;
                break;
            }

        if(newp){ // found a slot?
            // zero process structure
            memset(newp, 0, sizeof (struct p_tab));

            // select a unique pid 
            while(newp->p_pid == 0){
                if(nextpid++ > MAXPID)
                    nextpid = 20;
                newp->p_pid = nextpid;
                for(p=ptab; p<ptab+maxproc; p++)
                    if(p->p_status != P_EMPTY && p->p_pid == nextpid)
                        newp->p_pid = 0; // try again
            }
            newp->p_status = P_FORKING;
        }
    }

    newp->p_page = (newp-ptab) * 0x10 + 0x10 + OS_BANK;

    return newp;
}

/* This is the clock interrupt routine.   Its job is to increment the clock
 * counters, increment the tick count of the running process, and either
 * swap it out if it has been in long enough and is in user space or mark
 * it to be swapped out if in system space.  Also it decrements the alarm
 * clock of processes.
 * UZI180: This is CALLed from the actual Interrupting service
 *      routine (See MACHASM.ASZ).                             HFB
 * This must have no automatic or register variables.
 */

void timer_interrupt(void)
{
    /* Increment processes and global tick counters */
    if (udata.u_ptab->p_status == P_RUNNING)
        incrtick(udata.u_insys ? &udata.u_stime : &udata.u_utime);

    incrtick(&ticks);

    /* Do once-per-second things */
    if (++ticks_this_second == TICKSPERSEC)
    {
        static ptptr p;

        /* Update global time counters */
        ticks_this_second = 0;

        /* Update process alarm clocks */
        for(p=ptab; p < ptab+maxproc; ++p){
            if (p->p_alarm){
                p->p_alarm--;
                if(!p->p_alarm)
                    ssig(p, SIGALRM);
            }
        }
    }

    /* Check run time of current process */
    if ((++runticks >= udata.u_ptab->p_priority) && !udata.u_insys && inint){  // Time to swap out?
#ifdef DEBUG
        kprintf("[preempt]");
#endif
        udata.u_insys = true;
        udata.u_ptab->p_status = P_READY;
        swapout();
        udata.u_insys = false;               /* We have swapped back in */
    }
}

#ifdef DEBUG
char *syscall_name[] = {
       "exit",
       "open",
       "close",
       "creat",
       "mknod",
       "link",
       "unlink",
       "read",
       "write",
       "seek",
       "chdir",
       "sync",
       "access",
       "chmod",
       "chown",
       "stat",
       "fstat",
       "dup",
       "getpid",
       "getppid",
       "getuid",
       "umask",
       "getfsys",
       "execve",
       "wait",
       "setuid",
       "setgid",
       "time",
       "stime",
       "ioctl",
       "brk",
       "sbrk",
       "fork",
       "mount",
       "umount",
       "signal",
       "dup2",
       "pause",
       "alarm",
       "kill",
       "pipe",
       "getgid",
       "times",
       "utime",
       "geteuid",
       "getegid" 
};
#endif

// UZI system call handler
// we arrive here from syscall.s with the kernel paged in, using the kernel stack, interrupts enabled.
void unix_syscall(void)
{
    // NO LOCAL VARIABLES PLEASE
    udata.u_insys = true;
    udata.u_error = 0;

       /* UZI180 saves the Stack Pointer and arguments in the
        * Assembly Language Function handler in MACHASM.ASZ
        */
    if (udata.u_callno >= UZI_SYSCALL_COUNT){
        udata.u_error = EINVAL;
    }else{
#ifdef DEBUG
        kprintf ("\t\tpid %d: syscall %d\t%s(%x, %x, %x)\n", udata.u_ptab->p_pid, 
                udata.u_callno, syscall_name[udata.u_callno], udata.u_argn, udata.u_argn1, udata.u_argn2);
#endif
        // dispatch system call
        udata.u_retval = (*syscall_dispatch[udata.u_callno])();

#ifdef DEBUG
        kprintf ("\t\t\tpid %d: ret syscall %d, ret %x err %d\n", 
                udata.u_ptab->p_pid, udata.u_callno, udata.u_retval, udata.u_error);
#endif
    }

    chksigs();

    __critical {
        if (runticks >= udata.u_ptab->p_priority){       /* Time to swap out? */
            udata.u_ptab->p_status = P_READY;
            swapout();
        }
    }

    udata.u_insys = false;
}

/*-- sendsig (ptptr proc, int16 sig)
 *-- {
 *--     register ptptr p;

 *--     if (proc)
 *--         ssig (proc, sig);
 *--     else
 *--         for (p=ptab; p < ptab+maxproc; ++p)
 *--             if (p->p_status)
 *--                 ssig (p, sig);
 *-- }--*/

void sgrpsig(uint8_t tty, uint8_t sig)
{
    ptptr p;

    for (p=ptab; p < ptab+maxproc; ++p) {
        if (p->p_status && p->p_tty == tty)
            ssig(p, sig);
    }
}

/* This sees if the current process has any signals set, and handles them.
 */
void chksigs(void)
{
    uint8_t j;

    // any signals pending?
    if(!(udata.u_ptab->p_pending)){
        return;
    }

    // dispatch the lowest numbered signal
    for (j=1; j < NSIGS; ++j){
        if(!(sigmask(j) & udata.u_ptab->p_pending))
            continue;
        if (udata.u_sigvec[j] == SIG_DFL) {
#ifdef DEBUG
            kprintf("process terminated by signal: ");
#endif
            doexit (0, j);
        }

        if (udata.u_sigvec[j] != SIG_IGN) {
            // Arrange to call the user routine at return
            udata.u_ptab->p_pending &= ~sigmask(j);  // unset the bit
#ifdef DEBUG
            kprintf("about to process signal %d\n", j);
#endif
            udata.u_cursig = j;
        }
    }
}

void ssig(ptptr proc, uint8_t sig)
{
    uint16_t sigm;

    sigm = sigmask(sig);

    __critical {
        if(proc->p_status != P_EMPTY) {         /* Presumably was killed just now */
            if(!(proc->p_ignored & sigm)){
                switch(proc->p_status){
                    case P_PAUSE:
                    case P_WAIT:
                    case P_SLEEP:
                        proc->p_status = P_READY;
                    default:
                }

                proc->p_wait = NULL;
                proc->p_pending |= sigm;
            }
        }
    }
}

void doexit (int16_t val, int16_t val2)
{
    int16_t j;
    ptptr p;

#ifdef DEBUG
    kprintf("process %d exiting\n", udata.u_ptab->p_pid);
#endif

    __critical {
        for (j=0; j < UFTSIZE; ++j) {
            if (udata.u_files[j] != NO_FILE)
                doclose(j);
        }

        _sync();                         /* Not necessary, but a good idea. */

        udata.u_ptab->p_exitval = (val<<8) | (val2 & 0xff);

        i_deref(udata.u_cwd);

        /* Stash away child's execution tick counts in process table,
         * overlaying some no longer necessary stuff.
         */
        addtick(&udata.u_utime, &udata.u_cutime);
        addtick(&udata.u_stime, &udata.u_cstime);
        memcpy(&(udata.u_ptab->p_wait), &udata.u_utime, 2 * sizeof(uzitime_t));

        /* See if we have any children. Set child's parents to our parent */
        for (p=ptab; p < ptab+maxproc; ++p) {
            if (p->p_status  &&  p->p_pptr == udata.u_ptab  &&  p != udata.u_ptab)
                p->p_pptr = udata.u_ptab->p_pptr;
        }
    }

    /* Wake up a waiting parent, if any. */
    wakeup((char *)udata.u_ptab->p_pptr);

    udata.u_ptab->p_status = P_ZOMBIE;

    swapin(getproc());
    panic("doexit: won't exit");
}
