#include <version.h>
#include <kernel.h>
#include <timer.h>
#include <kdata.h>
#include <printf.h>
#include <devsd.h>

p_tab *init_process;
char *cmdline = (char*)0x81; // CP/M command line (we load at 0x88 so the first 7 bytes or so should survive)
char bootline[2];

void create_init(void);

void main(void)
{
    // setup state
    inint = false;
    udata.u_insys = true;
    maxproc = procmem/64;

    if (d_open(TTYDEV) != 0)
        panic("no tty");

    kprintf("UZI-socz80 version %s\n", software_version_string);
    kprintf("Copyright (c) 1998-2002 by H.F.Bower, D.Braun, S.Nitschke, H.Peraza\n");
    kprintf("Copyright (c) 2013 by Will Sowerbutts <will@sowerbutts.com>\n");

    // check we don't exceed the process table size limit
    if(maxproc > PTABSIZE){
        kprintf("WARNING: Increase PTABSIZE to %d to use available RAM\n", maxproc);
        maxproc = PTABSIZE;
    }
    
    kprintf("%dkB total RAM, %dkB available to processes (%d processes max)\n",
            ramsize, procmem, maxproc);

    bufinit();

    create_init();

    kprintf("Enabling interrupts ... ");
    __asm
        ei
    __endasm;
    kprintf("ok.\n");

    // initialise hardware devices
    sd_init();
    // TODO: let's add more hardware!!

    if (*cmdline) {
        while (*cmdline == ' ') ++cmdline;
        root_dev = *cmdline - '0';
    } else {
        kprintf("bootdev: ");
        udata.u_base = bootline;
        udata.u_sysio = 1;
        udata.u_count = 2;
        udata.u_euid = 0;		/* Always begin as superuser */

        cdread(TTYDEV); /* read root filesystem name from tty */
        root_dev = *bootline - '0';
    }

    /* Mount the root device */
    kprintf("Mounting root fs (root_dev=%d): ", root_dev);

    if (fmount(root_dev, NULLINODE))
        panic("no filesys");
    root = i_open(root_dev, ROOTINODE);
    if(!root)
        panic("no root");

    kprintf("OK\n");

    i_ref(udata.u_cwd = root);
    rdtime(&udata.u_time);

    kprintf("Starting /init\n");
    _execve();

    panic("no /init");           /* BIG Trouble if we Get Here!! */
}

void create_init(void)
{
    uint8_t *j;
    // userspace: 0x100+ 0   1   2   3   4   5   6   7   8   9   A   B   C
    const char arg[] = { '/','i','n','i','t',0,  0,  1,  1,  0,  0,  0,  0 };

    init_process = ptab_alloc();
    udata.u_ptab = init_process;
    newproc(init_process);
    init_process->p_status = P_RUNNING;
    init_process->p_fork_inf = udata.u_page;

    // wipe file table
    for (j=udata.u_files; j < (udata.u_files+UFTSIZE); ++j){
        *j = NO_FILE;
    }

    // Poke the execve arguments into user data space so _execve() can read them back
    uput(arg, PROGBASE, sizeof (arg));

    // Set up things to look like the process is calling _execve()
    udata.u_argn  = (uint16_t)PROGBASE;
    udata.u_argn1 = 0x107;       /* Arguments (just "/init") */
    udata.u_argn2 = 0x10b;       /* Environment (none) */
}

/* This checks to see if a user-supplied address is legitimate */
bool valaddr(char *base, uint16_t size)
{
    if (base < PROGBASE || base+size >= (char *)&udata)
    {
        udata.u_error = EFAULT;
        return false;
    }
    return true;
}

void panic(char *deathcry)
{
    const char *p;
    const char *panic_prefix = "\n\rPANIC: ";

    for(p=panic_prefix; *p; p++)
        putchar(*p);
    for(p=deathcry; *p; p++)
        putchar(*p);
    trap_monitor();
}

void idump(void)
{
    inoptr ip;
    ptptr pp;
    extern struct cinode i_tab[];

    kprintf("Err %d root %d\n", udata.u_error, root - i_tab);
    kprintf ("\tMAGIC\tDEV\tNUM\tMODE\tNLINK\t(DEV)\tREFS\tDIRTY\n");

    for (ip=i_tab; ip < i_tab+ITABSIZE; ++ip)
    {
        kprintf ("%d\t%d\t%d\t%u\t0%o\t",
                ip-i_tab, ip->c_magic,ip->c_dev, ip->c_num,
                ip->c_node.i_mode);
        kprintf ("%d\t%d\t%d\t%d\n",             /* line split for compiler */
                ip->c_node.i_nlink,ip->c_node.i_addr[0],
                ip->c_refs,ip->c_dirty);
        if(!ip->c_magic)
            break;
    }

    kprintf ("\n\tSTAT\tWAIT\tPID\tPPTR\tALARM\tPENDING\tIGNORED\tCHILD\n");
    for (pp=ptab; pp < ptab+PTABSIZE /*maxproc*/; ++pp)
    {
        if (pp->p_status == P_EMPTY) continue;
        kprintf ("%d\t%d\t0x%x\t%d\t",
                pp-ptab, pp->p_status, pp->p_wait,  pp->p_pid);
	kprintf ("%d\t%d\t0x%x\t0x%x\t%x\n",      /* line split for compiler */
                pp->p_pptr-ptab, pp->p_alarm, pp->p_pending,
                pp->p_ignored, pp->p_fork_inf);
    }   
    
    bufdump();

    kprintf ("insys %d ptab %d call %d cwd %d sp 0x%x\n",
            udata.u_insys,udata.u_ptab-ptab, udata.u_callno, udata.u_cwd-i_tab,
            udata.u_sp);
}
