#include <kernel.h>
#include <kdata.h>
#include <printf.h>
#include <timer.h>


/*******************************************
Getpid ()                        Function 18
============================================
Return Process ID Number (PID) to Caller.
********************************************/

int16_t _getpid(void)
{
    return (udata.u_ptab->p_pid);
}


/*******************************************
Getppid ()                       Function 19
============================================
Return Parent's Process ID Number (PPID).
********************************************/

int16_t _getppid(void)
{
    return (udata.u_ptab->p_pptr->p_pid);
}


/*******************************************
Getuid ()                        Function 20
============================================
Return User ID Number (UID) to Caller.
********************************************/

int16_t _getuid(void)
{
    return (udata.u_ptab->p_uid);
}


/*******************************************
Geteuid ()                       Function 44
============================================
Return Effective User ID Number (EUID).
********************************************/

int16_t _geteuid(void)
{
    return (udata.u_euid);
}


/*******************************************
Getgid ()                        Function 41
============================================
Return Group ID Number (GID) to Caller.
********************************************/

int16_t _getgid(void)
{
    return (udata.u_gid);
}


/*******************************************
Getegid ()                       Function 45
============================================
Return Effective Group ID Number (EGID).
********************************************/

int16_t _getegid(void)
{
    return (udata.u_egid);
}



/*******************************************
setuid (uid)                     Function 25        ?
int uid;
============================================
Set User ID Number (UID) of Process.  Must
be SuperUser or owner, else Error (EPERM).
********************************************/
#define uid (int)udata.u_argn

int16_t _setuid(void)
{
    if (super() || udata.u_ptab->p_uid == uid)
    {
        udata.u_ptab->p_uid = uid;
        udata.u_euid = uid;
        return (0);
    }
    udata.u_error = EPERM;
    return (-1);
}

#undef uid



/*******************************************
setgid (gid)                     Function 26        ?
int gid;
============================================
Set Group ID Number (GID).  Must be Super-
User or in Group to Set, else Error (EPERM).
********************************************/
#define gid (int16_t)udata.u_argn

int16_t _setgid(void)
{
    if (super() || udata.u_gid == gid)
    {
        udata.u_gid = gid;
        udata.u_egid = gid;
        return (0);
    }
    udata.u_error = EPERM;
    return (-1);
}

#undef gid



/*******************************************
time (tvec)                      Function 27
int tvec[];
============================================
Read Clock Time/Date to User Buffer.
********************************************/
#define tvec (int16_t *)udata.u_argn

int16_t _time(void)
{
    uputw (tod.t_time, tvec);
    uputw (tod.t_date, tvec+1);
    return (0);
}

#undef tvec



/*******************************************
stime (tvec)                     Function 28
int tvec[];
============================================
Set Clock Time (Currently unimplemented).
When active, must be SuperUser to Set Time.
********************************************/
#define tvec (int *)udata.u_argn

int16_t _stime(void)
{
    udata.u_error = EPERM;
    return (-1);
}

#undef tvec



/*******************************************
times (buf)                      Function 42        ?
char *buf;
********************************************/
#define buf (char *)udata.u_argn

int16_t _times(void)
{
    if(!(valaddr (buf, 6*sizeof(uzitime_t))))
        return (-1);

     /* If Destination is in Common Area (F000-FFFF is always in Context),
      * simply copy the data, otherwise perform Interbank "uput" to User Bank.
      */
    __critical {
        if (buf >= PROGTOP) {
            memcpy(buf, &udata.u_utime, 4*sizeof (uzitime_t));
            memcpy(buf + 4*sizeof (uzitime_t), &ticks, sizeof (uzitime_t));
        } else {
            uput(&udata.u_utime, buf, 4*sizeof (uzitime_t));
            uput(&ticks, buf + 4*sizeof (uzitime_t), sizeof (uzitime_t));
        }
    }
    return (0);
}

#undef buf



/* User's execve() call. All other flavors are library routines. */
/*******************************************
execve (name, argv, envp)        Function 23
char *name;
char *argv[];
char *envp[];
********************************************/
#define name (char *)udata.u_argn
#define argv (char **)udata.u_argn1
#define envp (char **)udata.u_argn2

int16_t _execve(void)
{
    inoptr ino, emu_ino;
    char     *buf;
    blkno_t  blk;
    char     **nargv;             /* In user space */
    char     **nenvp;             /* In user space */
    struct s_argblk *abuf, *ebuf;
    int      (**sigp)();
    int      argc;
    uint16_t emu_size, emu_copy;
    char     *progptr, *emu_ptr, *emu_base;

    if (!(ino = n_open(name, NULLINOPTR)))
        return (-1);

    if (!((getperm (ino) & OTH_EX) &&
          (ino->c_node.i_mode & F_REG) &&
          (ino->c_node.i_mode & (OWN_EX | OTH_EX | GRP_EX)) ) ) {
        udata.u_error = EACCES;
        goto nogood;
    }

    setftime(ino, A_TIME);

    /* Read in the first block of the new program */
    buf = bread(ino->c_dev, bmap(ino, 0, 1), 0);

    /****************************************
     * Get magic number into var magic
     * C3    : executable file no C/D sep.
     * 00FF  :     "        "  with C/D sep. (not supported in UZI180)
     * other : maybe shell script (nogood2)
     ****************************************/
    if ((*buf & 0xff) != EMAGIC) {
         udata.u_error = ENOEXEC;
         goto nogood2;
    }

    // Set Flag to indicate type of Executable (UZI or CP/M)
    if ((*(buf+3)=='U') && (*(buf+4)=='Z') && (*(buf+5)=='I')){
        emu_ino = 0;         // no emulation, thanks
    }else{
        // open the emulator code on disk
        emu_ino = kn_open(CPM_EMULATOR_FILENAME, NULLINOPTR);
        if(!emu_ino){
            kprintf("Cannot load emulator: %s\n", CPM_EMULATOR_FILENAME);
            udata.u_error = ENOEXEC;
            goto nogood2;
        }
    }

    /* Gather the arguments, and put them in temporary buffers. */
    abuf = (struct s_argblk *)tmpbuf();
    /* Put environment in another buffer. */
    ebuf = (struct s_argblk *)tmpbuf();

    // read args and environment from process memory
    if (rargs(argv, abuf) || rargs(envp, ebuf))
        goto nogood3;     /* SN */

    // setuid, setgid if executable requires it
    if (ino->c_node.i_mode & SET_UID)
        udata.u_euid = ino->c_node.i_uid;

    if (ino->c_node.i_mode & SET_GID)
        udata.u_egid = ino->c_node.i_gid;

    udata.u_ptab->p_fork_inf = 0;       /* Reset child of fork */

    /* We are definitely going to succeed with the exec,
     * so we can start writing over the old program
     */
    uput(buf, PROGBASE, 512);          /* Move 1st Block to user bank */
    brelse(buf);

    /* At this point, we are committed to reading in and
     * executing the program. */

    // Load the CP/M emulator if it is required
    if(emu_ino){
        emu_size = (512 * emu_ino->c_node.i_size.o_blkno) + emu_ino->c_node.i_size.o_offset;
        // round up to nearest multiple of 256 bytes, fit it in below PROGTOP
        emu_ptr = (char*)(PROGTOP - ((emu_size + 255) & 0xff00));
        emu_base = emu_ptr;
        blk = 0;

        while(emu_size){
            buf = bread(emu_ino->c_dev, bmap(emu_ino, blk, 1), 0); // read block
            emu_copy = min(512, emu_size);
            uput(buf, emu_ptr, emu_copy); // copy to userspace
            brelse((bufptr)buf); // release block
            // adjust pointers
            emu_ptr  += emu_copy;
            emu_size -= emu_copy;
            blk++;
        }
        // close emulator file
        i_deref(emu_ino);

        // zero out the remainder of memory between the top of the emulator and PROGTOP
        uzero(emu_ptr, PROGTOP - emu_ptr);
    }else{
        emu_base = PROGTOP;
    }

    // emu_base now points at the byte after the last byte the program can occupy

    // Read in the rest of the program, block by block
    progptr = PROGBASE+512; // we copied the first block already
    for(blk = 1; blk <= ino->c_node.i_size.o_blkno; ++blk)
    {
        buf = bread(ino->c_dev, bmap(ino, blk, 1), 0);
        uput(buf, progptr, 512);
        brelse((bufptr)buf);
        progptr += 512;
    }
    i_deref(ino);
    udata.u_break = (int)progptr; //  Set initial break for program

    // zero all remaining process memory above the last block loaded.
    uzero(progptr, emu_base - progptr);

    // Turn off caught signals
    for (sigp = udata.u_sigvec; sigp < (udata.u_sigvec+NSIGS); ++sigp)
        *sigp = SIG_DFL;

    // place the arguments, environment and stack at the top of userspace memory,
    // below the emulator (or commonmemory if no emulator in use)

    // Write back the arguments and the environment
    nargv = wargs(((char*)emu_base-2), abuf, &argc);
    nenvp = wargs((char *)(nargv), ebuf, NULL);

    // Fill in udata.u_name with Program invocation name
    uget((void*)ugetw(nargv), udata.u_name, 8);
    memcpy(udata.u_ptab->p_name, udata.u_name, 8);

    brelse(abuf);
    brelse(ebuf);

    // Shove argc and the address of argv just below envp
    uputw((uint16_t)nargv, nenvp - 1);
    uputw((uint16_t)argc, nenvp - 2);

    // Set stack pointer for the program
    udata.u_isp = nenvp - 2;

    // Start execution (never returns)
    if(emu_ino)
        doexec(emu_base);
    else
        doexec(PROGBASE);

    // tidy up in various failure modes:
nogood3:
    brelse(abuf);
    brelse(ebuf);
    i_deref(emu_ino);
nogood2:
    brelse(buf);
nogood:
    i_deref(ino);
    return (-1);
}

#undef name
#undef argv
#undef envp

/* SN    TODO      max (1024) 512 bytes for argv
               and max  512 bytes for environ
*/

bool rargs(char **userspace_argv, struct s_argblk *argbuf)
{
    char *ptr;    /* Address of base of arg strings in user space */
    uint8_t c;
    uint8_t *bufp;

    argbuf->a_argc = 0;              /* Store argc in argbuf */
    bufp = argbuf->a_buf;

    while (ptr = (char *)ugetw (userspace_argv++)) {
        ++(argbuf->a_argc);          /* Store argc in argbuf. */
        do {
            *bufp++ = c = ugetc (ptr++);
            if (bufp > argbuf->a_buf+500)
            {
                udata.u_error = E2BIG;
                return true; // failed
            }
        }
        while (c);
    }
    argbuf->a_arglen = bufp - argbuf->a_buf;    /*Store total string size. */
    return false; // success
}


char **wargs(char *ptr, struct s_argblk *argbuf, int  *cnt) // ptr is in userspace
{
    char **argv;         /* Address of users argv[], just below ptr */
    int  argc, arglen;
    char **argbase;
    char *sptr;

    sptr = argbuf->a_buf;

    /* Move them into the users address space, at the very top */
    ptr -= (arglen = argbuf->a_arglen);

    if(arglen){
        uput(sptr, ptr, arglen);
    }

    /* Set argv to point below the argument strings */
    argc = argbuf->a_argc;
    argbase = argv = (char **)ptr - (argc + 1);

    if(cnt){
        *cnt = argc;
    }

              /* Set each element of argv[] to point to its argument string */
    while(argc--)
    {
        uputw((uint16_t)ptr, argv++);
        if (argc) {
            do
                ++ptr;
            while (*sptr++);
        }
    }
    uputw(0, argv);      /*;;26Feb- Add Null Pointer to end of array*/
    return ((char **)argbase);
}



/*******************************************
brk (addr)                       Function 30
char *addr;
********************************************/
#define addr (char *)udata.u_argn

int16_t _brk(void)
{
    /* Don't allow break to be set past user's stack pointer */
    /*** St. Nitschke allow min. of 512 bytes for Stack ***/
    if (addr >= (char *)(udata.u_sp)-512) {
        udata.u_error = ENOMEM;
        return (-1);
    }
    udata.u_break = (unsigned)addr;
    return (0);
}

#undef addr



/*******************************************
sbrk (incr)                      Function 31
uint16_t incr;
********************************************/
#define incr (uint16_t)udata.u_argn

int16_t _sbrk(void)
{
    unsigned oldbrk;

    udata.u_argn += (oldbrk = udata.u_break);
    if ((unsigned)udata.u_argn < oldbrk)
        return (-1);
    if (_brk())             /* brk (udata.u_argn) */
        return (-1);

    return ((unsigned)oldbrk);
}

#undef incr



/*******************************************
wait (statloc)                   Function 24        ?
int *statloc;
********************************************/
#define statloc (int *)udata.u_argn

int16_t _wait(void)
{
    ptptr p;
    int retval;

    if (statloc && !valaddr((char*)statloc, sizeof(int))) {
        udata.u_error = EFAULT;
        return (-1);
    }

    /* See if we have any children. */
    for (p=ptab; p < ptab+maxproc; ++p) {
        if (p->p_status && p->p_pptr == udata.u_ptab && p != udata.u_ptab)
            goto ok;
    }
    udata.u_error = ECHILD;
    return (-1);
ok:
    /* Search for an exited child; */
    for (;;)
    {
        chksigs();
        if (udata.u_cursig) {
            udata.u_error = EINTR;
            return (-1);
        }
        for (p=ptab; p < ptab+maxproc; ++p) {
            if (p->p_status == P_ZOMBIE && p->p_pptr == udata.u_ptab) {
                if (statloc)
                    uputw(p->p_exitval, statloc);

                p->p_status = P_EMPTY;
                retval = p->p_pid;

                /* Add in child's time info.  It was stored on top */
                /* of p_wait in the childs process table entry. */
                addtick(&udata.u_cutime, p->p_wait);
                addtick(&udata.u_cstime, (uzitime_t*)(((char*)p->p_wait) + (sizeof(uzitime_t))));
                return (retval);
            }
        }
            /* Nothing yet, so wait */
        psleep (udata.u_ptab);
    }
}

#undef statloc



/*******************************************
_exit (val)                       Function 0
int16_t val;
********************************************/
#define val (int16_t)udata.u_argn

int16_t __exit(void)
{
    doexit(val, 0);
    return 0; // ... yeah. that might not happen.
}

#undef val

/*******************************************
fork ()                          Function 32
********************************************/

int16_t _fork(void)
{
    // allocate new process
    struct p_tab *new_process;
    int16_t r;

    new_process = ptab_alloc();
    if(!new_process){
        udata.u_error = EAGAIN;
        return -1;
    }

    __critical {
        // we're going to run our child process next, so mark this process as being ready to run
        udata.u_ptab->p_status = P_READY;
        // WRS: this is required for fork to operate correctly, apparently. can't see that it's used.
        new_process->p_fork_inf = udata.u_page;
        // kick off the new process (the bifurcation happens inside here, we returns in both 
        // the child and parent contexts)
        r = dofork(new_process);
    }

    return r;
}



/*******************************************
pause ()                         Function 37
********************************************/

int16_t _pause(void)
{
    psleep (0);
    udata.u_error = EINTR;
    return (-1);
}



/*******************************************
signal (sig, func)               Function 35        ?
int16_t sig;
int16_t (*func)();
********************************************/
#define sig (int16_t)udata.u_argn
#define func (int (*)())udata.u_argn1

int16_t _signal(void)
{
    int16_t retval;

/*    if (sig < 1 || sig == SIGKILL || sig >= NSIGS) */
    if (sig < 1 || sig >= NSIGS) {
        udata.u_error = EINVAL;
        goto nogood;
    }

    __critical{
        if (func == SIG_IGN){
            udata.u_ptab->p_ignored |= sigmask (sig);
        }else {
            if (func != SIG_DFL && !valaddr ((char *)func, 1)) {
                udata.u_error = EFAULT;
                goto nogood;
            }
            udata.u_ptab->p_ignored &= ~sigmask (sig);
        }
        retval = (int)udata.u_sigvec[sig];
        udata.u_sigvec[sig] = func;
    }
    return (retval);

nogood:
    return (-1);
}

#undef sig
#undef func



/*******************************************
kill (pid, sig)                  Function 39
int16_t pid;
int16_t sig;
********************************************/
#define pid (int16_t)udata.u_argn
#define sig (int16_t)udata.u_argn1

int16_t _kill(void)
{
    ptptr p;

    if (sig <= 0 || sig > 15)
    {
        udata.u_error = EINVAL;
        return (-1);
    }

    for (p=ptab; p < ptab+maxproc; ++p)
    {
            /*
             *     SN wcy permission check
             */
        if (p->p_pid == pid) {
            if (udata.u_ptab->p_uid == p->p_uid || super() ) {
                ssig (p, sig);
                return (0);
            } else {
                udata.u_error = EPERM;
                return (-1);
            }
        }
    }
    udata.u_error = ESRCH;
    return (-1);
}

#undef pid
#undef sig



/*******************************************
alarm (secs)                     Function 38        ?
uint16_t secs;
********************************************/
#define secs (int16_t)udata.u_argn

int16_t _alarm(void)
{
    int16_t retval;

    retval = udata.u_ptab->p_alarm;
    udata.u_ptab->p_alarm = secs;
    return (retval);
}

#undef secs
